/* ////////////////////////////////////////////////////////////

File Name: ioutil.cpp
Copyright (c) 2018 Soji Yamakawa.  All rights reserved.
http://www.ysflight.com

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

//////////////////////////////////////////////////////////// */

#include "ioutil.h"



std::vector <unsigned char> ReadBinaryFile(FILE *fp)
{
	std::vector <unsigned char> dat;

	fseek(fp,0,SEEK_END);
	auto fSize=ftell(fp);
	fseek(fp,0,SEEK_SET);

	dat.resize(fSize);
	fread(dat.data(),1,fSize,fp);

	return dat;
}

std::vector <unsigned char> ReadBinaryFile(const char fn[])
{
	std::vector <unsigned char> dat;
	FILE *fp=fopen(fn,"rb");
	if(nullptr!=fp)
	{
		dat=ReadBinaryFile(fp);
		fclose(fp);
	}
	return dat;
}

bool WriteBinaryFile(FILE *fp,const std::vector <unsigned char> &dat)
{
	if(dat.size()==fwrite(dat.data(),1,dat.size(),fp))
	{
		return true;
	}
	return false;
}

bool WriteBinaryFile(const char fn[],const std::vector <unsigned char> &dat)
{
	FILE *fp=fopen(fn,"wb");
	if(nullptr!=fp)
	{
		auto r=WriteBinaryFile(fp,dat);
		fclose(fp);
		return r;
	}
	return false;
}


////////////////////////////////////////////////////////////


Win32FloppyDisk::Win32FloppyDisk()
{
	hVol=INVALID_HANDLE_VALUE;
	CleanUp();
}
Win32FloppyDisk::~Win32FloppyDisk()
{
	CleanUp();
}
void Win32FloppyDisk::CleanUp(void)
{
	drvLetter=0;
	rwMode=NOT_OPEN;

	if(INVALID_HANDLE_VALUE!=hVol)
	{
		CloseHandle(hVol);
	}
	hVol=INVALID_HANDLE_VALUE;
}
void Win32FloppyDisk::PrintAndSetErrorMessage(const char str[]) const
{
	lastErrMsg=str;
	printf("%s\n",str);
}

void Win32FloppyDisk::PrintDeviceType(DEVICE_TYPE devType)
{
	switch(devType)
	{
	case FILE_DEVICE_8042_PORT:
		printf("FILE_DEVICE_8042_PORT\n");
		break;
	case FILE_DEVICE_ACPI:
		printf("FILE_DEVICE_ACPI\n");
		break;
	case FILE_DEVICE_BATTERY:
		printf("FILE_DEVICE_BATTERY\n");
		break;
	case FILE_DEVICE_BEEP:
		printf("FILE_DEVICE_BEEP\n");
		break;
	case FILE_DEVICE_BLUETOOTH:
		printf("FILE_DEVICE_BLUETOOTH\n");
		break;
	case FILE_DEVICE_BUS_EXTENDER:
		printf("FILE_DEVICE_BUS_EXTENDER\n");
		break;
	case FILE_DEVICE_CD_ROM:
		printf("FILE_DEVICE_CD_ROM\n");
		break;
	case FILE_DEVICE_CD_ROM_FILE_SYSTEM:
		printf("FILE_DEVICE_CD_ROM_FILE_SYSTEM\n");
		break;
	case FILE_DEVICE_CHANGER:
		printf("FILE_DEVICE_CHANGER\n");
		break;
	case FILE_DEVICE_CONTROLLER:
		printf("FILE_DEVICE_CONTROLLER\n");
		break;
	case FILE_DEVICE_CRYPT_PROVIDER:
		printf("FILE_DEVICE_CRYPT_PROVIDER\n");
		break;
	case FILE_DEVICE_DATALINK:
		printf("FILE_DEVICE_DATALINK\n");
		break;
	case FILE_DEVICE_DFS:
		printf("FILE_DEVICE_DFS\n");
		break;
	case FILE_DEVICE_DFS_FILE_SYSTEM:
		printf("FILE_DEVICE_DFS_FILE_SYSTEM\n");
		break;
	case FILE_DEVICE_DFS_VOLUME:
		printf("FILE_DEVICE_DFS_VOLUME\n");
		break;
	case FILE_DEVICE_DISK:
		printf("FILE_DEVICE_DISK\n");
		break;
	case FILE_DEVICE_DISK_FILE_SYSTEM:
		printf("FILE_DEVICE_DISK_FILE_SYSTEM\n");
		break;
	case FILE_DEVICE_DVD:
		printf("FILE_DEVICE_DVD\n");
		break;
	case FILE_DEVICE_FILE_SYSTEM:
		printf("FILE_DEVICE_FILE_SYSTEM\n");
		break;
	case FILE_DEVICE_FIPS:
		printf("FILE_DEVICE_FIPS\n");
		break;
	case FILE_DEVICE_FULLSCREEN_VIDEO:
		printf("FILE_DEVICE_FULLSCREEN_VIDEO\n");
		break;
	case FILE_DEVICE_INFINIBAND:
		printf("FILE_DEVICE_INFINIBAND\n");
		break;
	case FILE_DEVICE_INPORT_PORT:
		printf("FILE_DEVICE_INPORT_PORT\n");
		break;
	case FILE_DEVICE_KEYBOARD:
		printf("FILE_DEVICE_KEYBOARD\n");
		break;
	case FILE_DEVICE_KS:
		printf("FILE_DEVICE_KS\n");
		break;
	case FILE_DEVICE_KSEC:
		printf("FILE_DEVICE_KSEC\n");
		break;
	case FILE_DEVICE_MAILSLOT:
		printf("FILE_DEVICE_MAILSLOT\n");
		break;
	case FILE_DEVICE_MASS_STORAGE:
		printf("FILE_DEVICE_MASS_STORAGE\n");
		break;
	case FILE_DEVICE_MIDI_IN:
		printf("FILE_DEVICE_MIDI_IN\n");
		break;
	case FILE_DEVICE_MIDI_OUT:
		printf("FILE_DEVICE_MIDI_OUT\n");
		break;
	case FILE_DEVICE_MODEM:
		printf("FILE_DEVICE_MODEM\n");
		break;
	case FILE_DEVICE_MOUSE:
		printf("FILE_DEVICE_MOUSE\n");
		break;
	case FILE_DEVICE_MULTI_UNC_PROVIDER:
		printf("FILE_DEVICE_MULTI_UNC_PROVIDER\n");
		break;
	case FILE_DEVICE_NAMED_PIPE:
		printf("FILE_DEVICE_NAMED_PIPE\n");
		break;
	case FILE_DEVICE_NETWORK:
		printf("FILE_DEVICE_NETWORK\n");
		break;
	case FILE_DEVICE_NETWORK_BROWSER:
		printf("FILE_DEVICE_NETWORK_BROWSER\n");
		break;
	case FILE_DEVICE_NETWORK_FILE_SYSTEM:
		printf("FILE_DEVICE_NETWORK_FILE_SYSTEM\n");
		break;
	case FILE_DEVICE_NETWORK_REDIRECTOR:
		printf("FILE_DEVICE_NETWORK_REDIRECTOR\n");
		break;
	case FILE_DEVICE_NULL:
		printf("FILE_DEVICE_NULL\n");
		break;
	case FILE_DEVICE_PARALLEL_PORT:
		printf("FILE_DEVICE_PARALLEL_PORT\n");
		break;
	case FILE_DEVICE_PHYSICAL_NETCARD:
		printf("FILE_DEVICE_PHYSICAL_NETCARD\n");
		break;
	case FILE_DEVICE_PRINTER:
		printf("FILE_DEVICE_PRINTER\n");
		break;
	case FILE_DEVICE_SCANNER:
		printf("FILE_DEVICE_SCANNER\n");
		break;
	case FILE_DEVICE_SCREEN:
		printf("FILE_DEVICE_SCREEN\n");
		break;
	case FILE_DEVICE_SERENUM:
		printf("FILE_DEVICE_SERENUM\n");
		break;
	case FILE_DEVICE_SERIAL_MOUSE_PORT:
		printf("FILE_DEVICE_SERIAL_MOUSE_PORT\n");
		break;
	case FILE_DEVICE_SERIAL_PORT:
		printf("FILE_DEVICE_SERIAL_PORT\n");
		break;
	case FILE_DEVICE_SMARTCARD:
		printf("FILE_DEVICE_SMARTCARD\n");
		break;
	case FILE_DEVICE_SMB:
		printf("FILE_DEVICE_SMB\n");
		break;
	case FILE_DEVICE_SOUND:
		printf("FILE_DEVICE_SOUND\n");
		break;
	case FILE_DEVICE_STREAMS:
		printf("FILE_DEVICE_STREAMS\n");
		break;
	case FILE_DEVICE_TAPE:
		printf("FILE_DEVICE_TAPE\n");
		break;
	case FILE_DEVICE_TAPE_FILE_SYSTEM:
		printf("FILE_DEVICE_TAPE_FILE_SYSTEM\n");
		break;
	case FILE_DEVICE_TERMSRV:
		printf("FILE_DEVICE_TERMSRV\n");
		break;
	case FILE_DEVICE_TRANSPORT:
		printf("FILE_DEVICE_TRANSPORT\n");
		break;
	case FILE_DEVICE_UNKNOWN:
		printf("FILE_DEVICE_UNKNOWN\n");
		break;
	case FILE_DEVICE_VDM:
		printf("FILE_DEVICE_VDM\n");
		break;
	case FILE_DEVICE_VIDEO:
		printf("FILE_DEVICE_VIDEO\n");
		break;
	case FILE_DEVICE_VIRTUAL_DISK:
		printf("FILE_DEVICE_VIRTUAL_DISK\n");
		break;
	case FILE_DEVICE_VMBUS:
		printf("FILE_DEVICE_VMBUS\n");
		break;
	case FILE_DEVICE_WAVE_IN:
		printf("FILE_DEVICE_WAVE_IN\n");
		break;
	case FILE_DEVICE_WAVE_OUT:
		printf("FILE_DEVICE_WAVE_OUT\n");
		break;
	case FILE_DEVICE_WPD:
		printf("FILE_DEVICE_WPD\n");
		break;
	default:
		printf("Undefined device.\n");
		break;
	}
}

void Win32FloppyDisk::PrintMediaType(MEDIA_TYPE medType)
{
	switch(medType)
	{
	case Unknown:
		printf("Unknown media.\n");
		break;
	case RemovableMedia:
		printf("Removable Meida other than a floppy disk.\n");
		break;
	case FixedMedia:
		printf("Hard disk drive.\n");
		break;
	case F5_1Pt2_512:
		printf("Floppy disk: F5_1Pt2_512\n");
		break;
	case F3_1Pt44_512:
		printf("Floppy disk: F3_1Pt44_512\n");
		break;
	case F3_2Pt88_512:
		printf("Floppy disk: F3_2Pt88_512\n");
		break;
	case F3_20Pt8_512:
		printf("Floppy disk: F3_20Pt8_512\n");
		break;
	case F3_720_512:
		printf("Floppy disk: F3_720_512\n");
		break;
	case F5_360_512:
		printf("Floppy disk: F5_360_512\n");
		break;
	case F5_320_512:
		printf("Floppy disk: F5_320_512\n");
		break;
	case F5_320_1024:
		printf("Floppy disk: F5_320_1024\n");
		break;
	case F5_180_512:
		printf("Floppy disk: F5_180_512\n");
		break;
	case F5_160_512:
		printf("Floppy disk: F5_160_512\n");
		break;
	case F3_120M_512:
		printf("Floppy disk: F3_120M_512\n");
		break;
	case F3_640_512:
		printf("Floppy disk: F3_640_512\n");
		break;
	case F5_640_512:
		printf("Floppy disk: F5_640_512\n");
		break;
	case F5_720_512:
		printf("Floppy disk: F5_720_512\n");
		break;
	case F3_1Pt2_512:
		printf("Floppy disk: F3_1Pt2_512\n");
		break;
	case F3_1Pt23_1024:
		printf("Floppy disk: F3_1Pt23_1024\n");
		break;
	case F5_1Pt23_1024:
		printf("Floppy disk: F5_1Pt23_1024\n");
		break;
	case F3_128Mb_512:
		printf("Floppy disk: F3_128Mb_512\n");
		break;
	case F3_230Mb_512:
		printf("Floppy disk: F3_230Mb_512\n");
		break;
	case F8_256_128:
		printf("Floppy disk: F8_256_128\n");
		break;
	case F3_200Mb_512:
		printf("Floppy disk: F3_200Mb_512\n");
		break;
	case F3_240M_512:
		printf("Floppy disk: F3_240M_512\n");
		break;
	case F3_32M_512:
		printf("Floppy disk: F3_32M_512\n");
		break;
	default:
		printf("Undefined media.\n");
		break;
	}
}

MEDIA_TYPE Win32FloppyDisk::GetMediaType(void) const
{
	DWORD nByte;

	MEDIA_TYPE medType;
	if(TRUE==DeviceIoControl(hVol,IOCTL_STORAGE_GET_MEDIA_TYPES,NULL,0,&medType,sizeof(medType),&nByte,NULL))
	{
		PrintMediaType(medType);
	}
	else
	{
		medType=Unknown;
		printf("Failed to get media type.\n");
	}

	return medType;
}

//void Win32FloppyDisk::TestRead(void) const
//{
//	static unsigned char dat[2048*1024];
//	DWORD nByte;
//
//	if(TRUE!=ReadFile(hVol,dat,1024 /*sizeof(dat)*/,&nByte,NULL))
//	{
//		printf("Failed to read.\n");
//		if(0<nByte)
//		{
//			printf("Read %d bytes.  Saving anyway.\n",nByte);
//			FILE *fp=fopen("xyz.xdf","wb");
//			fwrite(dat,1,nByte,fp);
//			fclose(fp);
//		}
//	}
//	else
//	{
//		printf("Read %d bytes.\n",nByte);
//		FILE *fp=fopen("xyz.xdf","wb");
//		fwrite(dat,1,nByte,fp);
//		fclose(fp);
//	}
//}

bool Win32FloppyDisk::OpenForRead(char drvLetter)
{
	if(NOT_OPEN!=rwMode)
	{
		PrintAndSetErrorMessage("Already Open.");
		return false;
	}

	if('a'<=drvLetter && drvLetter<='z')
	{
		drvLetter=drvLetter-'a'+'A';
	}
	if(drvLetter<'A' || 'Z'<drvLetter)
	{
		char msg[256];
		sprintf(msg,"Drive Letter %c is invalid.",drvLetter);
		PrintAndSetErrorMessage(msg);
		return false;
	}

	this->drvLetter=drvLetter;

	CleanUp();  // Point of no return.

	char drvFn[256];
	sprintf(drvFn,"\\\\.\\%c:",drvLetter);
	printf("Opening Drive: %s\n",drvFn);
	hVol=CreateFile(drvFn,GENERIC_READ,FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,OPEN_EXISTING,0,NULL);
	if(INVALID_HANDLE_VALUE==hVol)
	{
		char msg[256];
		sprintf(msg,"Failed to open drive %c.",drvLetter);
		PrintAndSetErrorMessage(msg);
		CleanUp();
		return false;
	}

	printf("Drive opened.\n");

	auto medType=GetMediaType();
	PrintMediaType(medType);

	return true;
}

bool Win32FloppyDisk::OpenForWrite(char drvLetter)
{
	if(NOT_OPEN!=rwMode)
	{
		PrintAndSetErrorMessage("Already Open.");
		return false;
	}

	if('a'<=drvLetter && drvLetter<='z')
	{
		drvLetter=drvLetter-'a'+'A';
	}
	if(drvLetter<'A' || 'Z'<drvLetter)
	{
		char msg[256];
		sprintf(msg,"Drive Letter %c is invalid.",drvLetter);
		PrintAndSetErrorMessage(msg);
		return false;
	}

	this->drvLetter=drvLetter;

	CleanUp();  // Point of no return.

	char drvFn[256];
	sprintf(drvFn,"\\\\.\\%c:",drvLetter);
	printf("Opening Drive: %s\n",drvFn);
	hVol=CreateFile(drvFn,GENERIC_WRITE,FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,OPEN_EXISTING,0,NULL);
	if(INVALID_HANDLE_VALUE==hVol)
	{
		char msg[256];
		sprintf(msg,"Failed to open drive %c.",drvLetter);
		PrintAndSetErrorMessage(msg);
		CleanUp();
		return false;
	}

	printf("Drive opened.\n");

	auto medType=GetMediaType();
	PrintMediaType(medType);

	return true;
}


bool Win32FloppyDisk::WriteImage(long long int writeSize,const unsigned char dat[])
{
	bool res=true;
	DWORD nByte;
	DeviceIoControl(hVol,FSCTL_LOCK_VOLUME,NULL,0,NULL,0,&nByte,NULL);
	if(TRUE!=WriteFile(hVol,dat,writeSize,&nByte,NULL))
	{
		PrintAndSetErrorMessage("Failed to write.\n");
		res=false;
	}
	else
	{
		printf("Wrote %d bytes.\n",nByte);
		res=true;
	}
	DeviceIoControl(hVol,FSCTL_UNLOCK_VOLUME,NULL,0,NULL,0,&nByte,NULL);
	return res;
}

bool Win32FloppyDisk::Format(int track,int side,int gapLength,int sectorSize,int numSecPerTrack)
{
	DWORD nByte;
	const int maxNumSecPerTrack=32;
	unsigned char buf[sizeof(FORMAT_EX_PARAMETERS)+maxNumSecPerTrack*sizeof(WORD)];
	FORMAT_EX_PARAMETERS &param=*((FORMAT_EX_PARAMETERS *)buf);
	param.MediaType=F3_1Pt23_1024;
	param.StartCylinderNumber=track;
	param.EndCylinderNumber=track;
	param.StartHeadNumber=side;
	param.EndHeadNumber=side;
	param.FormatGapLength=gapLength;
	param.SectorsPerTrack=numSecPerTrack;
	for(int i=1; i<=numSecPerTrack; ++i)
	{
		param.SectorNumber[i-1]=i;
	}
	if(TRUE!=DeviceIoControl(hVol,IOCTL_DISK_FORMAT_TRACKS_EX,
	             &param,sizeof(FORMAT_EX_PARAMETERS)+sizeof(WORD)*(numSecPerTrack-1),
	             NULL,0,
	             &nByte,NULL))
	{
		char msg[256];
		sprintf(msg,"Failed to format track %d side %d\n",track,side);
		PrintAndSetErrorMessage(msg);
		return false;
	}
	return true;
}


void Win32FloppyDisk::Close(void)
{
	CleanUp();
}
