/* ////////////////////////////////////////////////////////////

File Name: ioutil.h
Copyright (c) 2018 Soji Yamakawa.  All rights reserved.
http://www.ysflight.com

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

//////////////////////////////////////////////////////////// */

#ifndef IOUTIL_IS_INCLUDED
#define IOUTIL_IS_INCLUDED
/* { */

#include <windows.h>
#include <fileapi.h>
#include <stdio.h>

#include <vector>
#include <string>


std::vector <unsigned char> ReadBinaryFile(FILE *fp);
std::vector <unsigned char> ReadBinaryFile(const char fn[]);
bool WriteBinaryFile(FILE *fp,const std::vector <unsigned char> &dat);
bool WriteBinaryFile(const char fn[],const std::vector <unsigned char> &dat);


class Win32FloppyDisk
{
public:
	enum OPENMODE
	{
		NOT_OPEN,
		READING_FROM_PHYS,
		WRITING_TO_PHYS
	};

private:
	char drvLetter;
	OPENMODE rwMode;
	HANDLE hVol;

public:
	Win32FloppyDisk();
	~Win32FloppyDisk();
	void CleanUp(void);

private:
	void PrintAndSetErrorMessage(const char str[]) const;

	static void PrintDeviceType(DEVICE_TYPE devType);
	static void PrintMediaType(MEDIA_TYPE medType);

	MEDIA_TYPE GetMediaType(void) const;

public:
	mutable std::string lastErrMsg;
	bool OpenForRead(char drvLetter);
	bool OpenForWrite(char drvLetter);

	bool WriteImage(long long int nByte,const unsigned char dat[]);
	bool Format(int track,int side,int gapLength,int sectorSize,int numSecPerTrack);

	void Close(void);
};


/* } */
#endif
